import { __assign, __spreadArray } from 'tslib';
import { jsx } from 'react/jsx-runtime';
import { useMemo, useReducer, useCallback } from 'react';
import { SELECTION_TYPE, SelectionCell } from './Selection.js';
import useRunOnceWhenTruthy from '../../hooks/useRunOnceWhenTruthy.js';
import { ManageColumns } from './ManageColumns.js';
import { Checkbox } from '../Checkbox/index.js';
import { css } from '@emotion/css';
import { getStickyClass } from '../../utils/table.js';
import classNames from 'classnames';
import _orderBy from 'lodash/orderBy';
import _filter from 'lodash/filter';
import _forEach from 'lodash/forEach';
import _isString from 'lodash/isString';
import _get from 'lodash/get';
import _isObject from 'lodash/isObject';
import _find from 'lodash/find';
import _chunk from 'lodash/chunk';
import _isArray from 'lodash/isArray';
import _uniqBy from 'lodash/uniqBy';
import _compact from 'lodash/compact';
import _map from 'lodash/map';
import _isUndefined from 'lodash/isUndefined';
import _pickBy from 'lodash/pickBy';
import _size from 'lodash/size';
import _values from 'lodash/values';

var noRightPaddingClass = css({
  '&&&&': {
    paddingRight: 0,
    '--display-cell-right-border': 'none'
  }
});
var noFilterState = {
  columns: {},
  count: 0,
  textFilter: ''
};
var sort = function (sortState, columnsMap, rows) {
  return (sortState === null || sortState === void 0 ? void 0 : sortState.column) && (sortState === null || sortState === void 0 ? void 0 : sortState.sortOrder) ? _orderBy(rows, [columnsMap[sortState.column].accessor], [sortState.sortOrder]) : rows;
};
var filterFunc = function (filterState, rows, columns, columnsMap, additionalSearchKeys) {
  var textFilter = filterState.textFilter;
  return _filter(rows, function (row) {
    var lowerCaseTextFilter = textFilter && textFilter.toLowerCase();
    var isTextFilterMatch = !textFilter;
    if (textFilter) {
      _forEach(columns, function (column) {
        var accessor = column.accessor;
        var accessorForTextFilter = column.accessorForTextFilter;
        var value = accessorForTextFilter ? _get(row, accessorForTextFilter) : _get(row, accessor);
        if (_isString(value) && value.toLowerCase().includes(lowerCaseTextFilter)) {
          isTextFilterMatch = true;
          return false;
        } else if (_isArray(value) && value.join(', ').toLowerCase().includes(lowerCaseTextFilter)) {
          isTextFilterMatch = true;
          return false;
        }
      });
      additionalSearchKeys && _forEach(additionalSearchKeys, function (searchKey) {
        var value = _get(row, searchKey);
        if (_isString(value) && value.toLowerCase().includes(lowerCaseTextFilter)) {
          isTextFilterMatch = true;
          return false;
        }
      });
    }
    var isMultiFilterMatch = true;
    _forEach(filterState.columns, function (filter, key) {
      if (_isObject(filter) && filter.activeCount > 0) {
        var accessor = columnsMap[key].accessor;
        var rowAccessor = _get(row, accessor);
        if (_isArray(rowAccessor)) {
          var existsOne = _find(rowAccessor, function (value) {
            return filter.values[value];
          });
          if (!existsOne) {
            isMultiFilterMatch = false;
            return false;
          }
        } else {
          if (!filter.values[rowAccessor]) {
            isMultiFilterMatch = false;
            return false;
          }
        }
      }
    });
    return isTextFilterMatch && isMultiFilterMatch;
  });
};
var ACTIONS = {
  TOGGLE_SORT: 'TOGGLE_SORT',
  SET_FILTER: 'SET_FILTER',
  RESET_FILTERS: 'RESET_FILTERS',
  TOGGLE_ROW_SELECTION: 'TOGGLE_ROW_SELECTION',
  TOGGLE_SELECT_ALL: 'TOGGLE_SELECT_ALL',
  UPDATE_TEXT_FILTER: 'UPDATE_TEXT_FILTER',
  GO_TO_PAGE: 'GO_TO_PAGE',
  UPDATE_ROW: 'UPDATE_ROW',
  UPDATE_COLUMN_STATE: 'UPDATE_COLUMN_STATE'
};
function init(initialState) {
  var sortState = null;
  if (initialState.isSorting && initialState.columns) {
    sortState = {
      column: _isString(initialState.columns[0]) ? initialState.columns[0] : initialState.columns[0].id || '0',
      sortOrder: 'asc'
    };
  }
  var initialPageState = initialState.pageSize ? {
    pageIndex: 0
  } : undefined;
  return __assign(__assign({}, initialState), {
    sortState: initialState.initialSortState || sortState,
    initialSortState: initialState.initialSortState || sortState,
    filterState: initialState.initialFilterState || noFilterState,
    initialFilterState: initialState.initialFilterState || noFilterState,
    columnsState: initialState.initialColumnState || {},
    rowsState: {},
    selectionState: initialState.selectionType !== SELECTION_TYPE.NONE ? {
      rows: {},
      count: 0
    } : undefined,
    selectionType: initialState.selectionType,
    pageState: initialPageState,
    initialPageState: initialPageState
  });
}
function reducer(state, action) {
  var _a, _b, _c;
  var _d, _e, _f;
  var type = action.type,
    payload = action.payload;
  switch (type) {
    case ACTIONS.TOGGLE_SORT:
      {
        var id = payload.id;
        var newSortState = {};
        var currentSortOrder = (_d = state === null || state === void 0 ? void 0 : state.sortState) === null || _d === void 0 ? void 0 : _d.sortOrder;
        var currentSortColumn = (_e = state === null || state === void 0 ? void 0 : state.sortState) === null || _e === void 0 ? void 0 : _e.column;
        if (currentSortColumn === id) {
          if (currentSortOrder === 'asc') {
            newSortState.sortOrder = 'desc';
            newSortState.column = id;
          } else if (currentSortColumn !== ((_f = state.initialSortState) === null || _f === void 0 ? void 0 : _f.column)) {
            newSortState = state.initialSortState;
          } else {
            newSortState.sortOrder = 'asc';
            newSortState.column = id;
          }
        } else {
          newSortState.sortOrder = 'asc';
          newSortState.column = id;
        }
        return __assign(__assign({}, state), {
          sortState: newSortState,
          pageState: state.initialPageState
        });
      }
    case ACTIONS.SET_FILTER:
      {
        var id = payload.id,
          values = payload.values;
        var newValues = _pickBy(values);
        var newColumnFilter = {
          values: newValues,
          valuesArray: _values(values),
          activeCount: _size(newValues)
        };
        var columns = __assign(__assign({}, state.filterState.columns), (_a = {}, _a[id] = newColumnFilter, _a));
        var count_1 = 0;
        _forEach(columns, function (columnsState) {
          if (columnsState.activeCount && columnsState.activeCount > 0) {
            count_1 = count_1 + 1;
          }
        });
        return __assign(__assign({}, state), {
          filterState: __assign(__assign({}, state === null || state === void 0 ? void 0 : state.filterState), {
            columns: columns,
            count: count_1
          }),
          pageState: state.initialPageState
        });
      }
    case ACTIONS.RESET_FILTERS:
      {
        return __assign(__assign({}, state), {
          filterState: noFilterState,
          pageState: state.initialPageState
        });
      }
    case ACTIONS.TOGGLE_ROW_SELECTION:
      {
        var id = payload.id,
          value = payload.value;
        var count = state.selectionState.count;
        var updatedValue = value;
        var oldSelectedRows = state.selectionState.rows;
        if (state.selectionType === SELECTION_TYPE.MULTIPLE) {
          var currentValue = state.selectionState.rows[id] || false;
          if (currentValue && !value) {
            count = count - 1;
          } else if (!currentValue && value) {
            count = count + 1;
          }
        }
        if (state.selectionType === SELECTION_TYPE.SINGULAR) {
          count = 1;
          updatedValue = true;
          oldSelectedRows = {};
        }
        return __assign(__assign({}, state), {
          selectionState: {
            count: count,
            rows: __assign(__assign({}, oldSelectedRows), (_b = {}, _b[id] = updatedValue, _b))
          }
        });
      }
    case ACTIONS.TOGGLE_SELECT_ALL:
      {
        var value_1 = payload.value,
          rows = payload.rows;
        var selectionState_1 = {
          rows: {},
          count: 0,
          allSelected: value_1
        };
        if (value_1) {
          var notDisabledRows = rows.filter(function (row) {
            var _a;
            return !((_a = row === null || row === void 0 ? void 0 : row.cellProps) === null || _a === void 0 ? void 0 : _a.isDisabled);
          });
          _forEach(notDisabledRows, function (row) {
            var _a;
            if ((row === null || row === void 0 ? void 0 : row.id) && !((_a = row === null || row === void 0 ? void 0 : row.cellProps) === null || _a === void 0 ? void 0 : _a.isDisabled)) {
              selectionState_1.rows[row.id] = value_1;
            }
          });
          selectionState_1.count = notDisabledRows.length;
        }
        return __assign(__assign({}, state), {
          selectionState: selectionState_1
        });
      }
    case ACTIONS.UPDATE_TEXT_FILTER:
      {
        var value = payload.value;
        return __assign(__assign({}, state), {
          filterState: __assign(__assign({}, state === null || state === void 0 ? void 0 : state.filterState), {
            textFilter: value
          }),
          pageState: state.initialPageState
        });
      }
    case ACTIONS.GO_TO_PAGE:
      {
        return __assign(__assign({}, state), {
          pageState: {
            pageIndex: payload.index
          }
        });
      }
    case ACTIONS.UPDATE_ROW:
      {
        var id = payload.id,
          newState = payload.state;
        var currentState = state.rowsState[id] || {};
        return __assign(__assign({}, state), {
          rowsState: __assign(__assign({}, state.rowsState), (_c = {}, _c[id] = __assign(__assign({}, currentState), newState), _c))
        });
      }
    case ACTIONS.UPDATE_COLUMN_STATE:
      {
        return __assign(__assign({}, state), {
          columnsState: payload.state
        });
      }
  }
  return state;
}
var getFilterOptions = function (data, column) {
  var accessor = column.accessor,
    renderFilterPanelLabel = column.renderFilterPanelLabel;
  return !data ? [] : _orderBy(_uniqBy(_compact(_map(data, function (row) {
    var value = _get(row, accessor, null);
    if (!value) return null;
    // renderFilterPanelLabel is a simpler version of renderCell, for filters
    return {
      value: value,
      label: renderFilterPanelLabel ? renderFilterPanelLabel(value) : value
    };
  })), 'value'), [function (_a) {
    var value = _a.value;
    return typeof value === 'string' ? value.toLowerCase() : JSON.stringify(value);
  }], ['asc']);
};
var useTable = function (_a) {
  var _b;
  var isSorting = _a.isSorting,
    _c = _a.selectionType,
    selectionType = _c === void 0 ? 'none' : _c,
    columns = _a.columns,
    data = _a.rows,
    _d = _a.isExternalSort,
    isExternalSort = _d === void 0 ? false : _d,
    _e = _a.isExternalFilter,
    isExternalFilter = _e === void 0 ? false : _e,
    _f = _a.initialSortState,
    initialSortState = _f === void 0 ? null : _f,
    _g = _a.initialFilterState,
    initialFilterState = _g === void 0 ? noFilterState : _g,
    initialColumnState = _a.initialColumnState,
    _h = _a.pageSize,
    pageSize = _h === void 0 ? 50 : _h,
    defaultSelectedRows = _a.defaultSelectedRows,
    additionalSearchKeys = _a.additionalSearchKeys,
    _j = _a.isManagedColumns,
    isManagedColumns = _j === void 0 ? false : _j,
    _k = _a.isHorizontalScroll,
    isHorizontalScroll = _k === void 0 ? false : _k,
    _l = _a.isLazyLoading,
    isLazyLoading = _l === void 0 ? false : _l,
    _m = _a.isSelectAllDisabled,
    isSelectAllDisabled = _m === void 0 ? false : _m;
  var _o = useMemo(function () {
      var columnsMap = {};
      var pColumns = columns === null || columns === void 0 ? void 0 : columns.map(function (column, index) {
        var pColumn = __assign(__assign({}, column), {
          id: _isUndefined(column.id) ? index.toString() : column.id
        });
        columnsMap[pColumn.id] = pColumn;
        return pColumn;
      });
      return {
        pColumns: pColumns,
        columnsMap: columnsMap
      };
    }, [columns]),
    pColumns = _o.pColumns,
    columnsMap = _o.columnsMap;
  var _p = useReducer(reducer, init({
      columns: columns,
      isSorting: isSorting,
      selectionType: selectionType,
      initialSortState: initialSortState,
      initialFilterState: initialFilterState,
      initialColumnState: initialColumnState,
      pageSize: pageSize
    })),
    table = _p[0],
    dis = _p[1];
  var dispatch = dis;
  var toggleRowSelection = useCallback(function (id) {
    return function (value) {
      dispatch({
        type: ACTIONS.TOGGLE_ROW_SELECTION,
        payload: {
          id: id,
          value: value
        }
      });
    };
  }, [dispatch]);
  var processedRows = useMemo(function () {
    return _map(data, function (row, index) {
      var id = _isUndefined(row.id) ? index.toString() : row.id;
      return __assign(__assign({}, row), {
        id: id
      });
    });
  }, [data, table, selectionType]);
  var shownColumns = useMemo(function () {
    return pColumns.filter(function (column) {
      var _a;
      return !((_a = table.columnsState[column.id]) === null || _a === void 0 ? void 0 : _a.isHidden);
    });
  }, [pColumns, table.columnsState]);
  var toggleSort = useCallback(function (_a) {
    var id = _a.id;
    dispatch({
      type: ACTIONS.TOGGLE_SORT,
      payload: {
        id: id
      }
    });
  }, [dispatch]);
  var updateFilterState = useCallback(function (payload) {
    dispatch({
      type: ACTIONS.SET_FILTER,
      payload: payload
    });
  }, [dispatch]);
  var resetFilters = useCallback(function () {
    dispatch({
      type: ACTIONS.RESET_FILTERS
    });
  }, [dispatch]);
  var updateRowState = useCallback(function (id) {
    return function (state) {
      dispatch({
        type: ACTIONS.UPDATE_ROW,
        payload: {
          id: id,
          state: state
        }
      });
    };
  }, [dispatch]);
  var updateTextFilter = useCallback(function (value) {
    dispatch({
      type: ACTIONS.UPDATE_TEXT_FILTER,
      payload: {
        value: value
      }
    });
  }, [dispatch]);
  var filteredRows = useMemo(function () {
    var _a, _b;
    if ((((_a = table.filterState) === null || _a === void 0 ? void 0 : _a.count) > 0 || ((_b = table.filterState) === null || _b === void 0 ? void 0 : _b.textFilter)) && !isExternalFilter) {
      return filterFunc(table.filterState, processedRows, columns, columnsMap, additionalSearchKeys);
    } else {
      return processedRows;
    }
  }, [table.filterState, processedRows, isExternalFilter, columns, columnsMap]);
  var columnsWithFilterOptions = useMemo(function () {
    return _map(shownColumns, function (col) {
      return __assign(__assign({}, col), {
        filterOptions: col.filterOptions === 'auto' ? getFilterOptions(filteredRows, col) : col.filterOptions
      });
    });
  }, [filteredRows, shownColumns]);
  var organizedRows = useMemo(function () {
    if (!isExternalSort) {
      return sort(table.sortState, columnsMap, filteredRows);
    } else {
      return filteredRows;
    }
  }, [isSorting, isExternalSort, filteredRows, table.sortState, columnsMap]);
  var selectRows = useCallback(function (rows, value) {
    dispatch({
      type: ACTIONS.TOGGLE_SELECT_ALL,
      payload: {
        value: value,
        rows: rows
      }
    });
  }, [dispatch]);
  var pageChunks = useMemo(function () {
    if (pageSize) {
      return _chunk(organizedRows, pageSize);
    } else {
      return null;
    }
  }, [organizedRows, pageSize]);
  var pagination = useMemo(function () {
    var _a;
    if (pageChunks) {
      var pageIndex = ((_a = table === null || table === void 0 ? void 0 : table.pageState) === null || _a === void 0 ? void 0 : _a.pageIndex) || 0;
      var pageCount_1 = pageChunks.length;
      var gotoPage = function (i) {
        if (i >= 0 && i < pageCount_1) {
          dispatch({
            type: ACTIONS.GO_TO_PAGE,
            payload: {
              index: i
            }
          });
        }
      };
      return {
        pageIndex: pageIndex,
        pageRows: pageChunks[pageIndex],
        pageCount: pageCount_1,
        gotoPage: gotoPage
      };
    } else {
      return null;
    }
  }, [pageChunks, table.pageState, dispatch]);
  var processedColumns = useMemo(function () {
    return columnsWithFilterOptions.map(function (column) {
      var _a, _b, _c;
      return __assign(__assign({}, column), {
        toggleSort: function () {
          return toggleSort({
            id: column.id
          });
        },
        sortState: column.id === ((_a = table.sortState) === null || _a === void 0 ? void 0 : _a.column) ? (_b = table.sortState) === null || _b === void 0 ? void 0 : _b.sortOrder : null,
        filterState: (_c = table.filterState) === null || _c === void 0 ? void 0 : _c.columns[column.id],
        updateColumnFilter: function (values) {
          return updateFilterState({
            id: column.id,
            values: values
          });
        }
      });
    });
  }, [toggleSort, table.sortState, updateFilterState, table.filterState, columnsWithFilterOptions]);
  var columnsWithSelection = useMemo(function () {
    var _a;
    var isAllSelected = (organizedRows === null || organizedRows === void 0 ? void 0 : organizedRows.every(function (row) {
      var _a, _b;
      return ((_a = table.selectionState) === null || _a === void 0 ? void 0 : _a.rows[row.id]) || ((_b = row === null || row === void 0 ? void 0 : row.cellProps) === null || _b === void 0 ? void 0 : _b.isDisabled);
    })) && (organizedRows === null || organizedRows === void 0 ? void 0 : organizedRows.length) > 0;
    var isAllPageRowsDisabled = (_a = pagination === null || pagination === void 0 ? void 0 : pagination.pageRows) === null || _a === void 0 ? void 0 : _a.every(function (row) {
      var _a;
      return (_a = row === null || row === void 0 ? void 0 : row.cellProps) === null || _a === void 0 ? void 0 : _a.isDisabled;
    });
    var isAllRowsDisabled = organizedRows === null || organizedRows === void 0 ? void 0 : organizedRows.every(function (row) {
      var _a;
      return (_a = row === null || row === void 0 ? void 0 : row.cellProps) === null || _a === void 0 ? void 0 : _a.isDisabled;
    });
    return table.selectionState ? __spreadArray([{
      Header: function () {
        var _a;
        if (selectionType === SELECTION_TYPE.MULTIPLE) {
          if (pagination.pageCount > 1) {
            return jsx(Checkbox, {
              isChecked: isAllSelected,
              isDisabled: isAllRowsDisabled || isSelectAllDisabled,
              isPartial: table.selectionState.count > 0 && !isAllSelected,
              variant: 'tableHeader',
              menuOptions: [{
                label: 'Select this page',
                isDisabled: isAllPageRowsDisabled,
                onClick: function () {
                  selectRows(pagination.pageRows, true);
                }
              }, {
                label: 'Select all pages',
                isActive: isAllSelected,
                onClick: function () {
                  selectRows(organizedRows, true);
                }
              }, {
                label: 'Clear all',
                isDisabled: ((_a = table === null || table === void 0 ? void 0 : table.selectionState) === null || _a === void 0 ? void 0 : _a.count) === 0,
                onClick: function () {
                  selectRows(organizedRows, false);
                }
              }]
            });
          } else {
            return jsx(Checkbox, {
              variant: 'tableCheckbox',
              isDisabled: isAllPageRowsDisabled || isSelectAllDisabled,
              isPartial: table.selectionState.count > 0 && !isAllSelected && (organizedRows === null || organizedRows === void 0 ? void 0 : organizedRows.length) > 0,
              isChecked: isAllSelected,
              onChange: function () {
                selectRows(organizedRows, !isAllSelected);
              }
            });
          }
        }
      },
      accessor: '',
      id: 'selectionColumn',
      className: noRightPaddingClass,
      isSticky: true,
      width: '72px',
      renderCell: function (value, row) {
        var _a, _b, _c;
        var isSelected = ((_a = table.selectionState) === null || _a === void 0 ? void 0 : _a.rows[row.id]) || false;
        return jsx(SelectionCell, __assign({
          selectionType: selectionType,
          isSelected: isSelected,
          isDisabled: ((_b = row.cellProps) === null || _b === void 0 ? void 0 : _b.isDisabled) || false,
          onChange: function () {
            return toggleRowSelection(row.id)(!isSelected);
          }
        }, (_c = row === null || row === void 0 ? void 0 : row.cellProps) === null || _c === void 0 ? void 0 : _c.selectionProps));
      }
    }], processedColumns, true) : processedColumns;
  }, [processedColumns, organizedRows, pagination === null || pagination === void 0 ? void 0 : pagination.pageRows, pagination === null || pagination === void 0 ? void 0 : pagination.pageCount, (_b = table.selectionState) === null || _b === void 0 ? void 0 : _b.count]);
  var updateColumnState = useCallback(function (state) {
    dispatch({
      type: ACTIONS.UPDATE_COLUMN_STATE,
      payload: {
        state: state
      }
    });
  }, [dispatch]);
  var columnsWithManageColumns = useMemo(function () {
    return isManagedColumns ? __spreadArray(__spreadArray([], columnsWithSelection, true), [{
      Header: function () {
        return jsx(ManageColumns, {
          allColumns: pColumns,
          columnsState: table.columnsState,
          updateColumnState: updateColumnState
        });
      },
      accessor: '',
      isSticky: true,
      width: '60px'
    }], false) : columnsWithSelection;
  }, [columns, table.columnsState, updateColumnState, columnsWithSelection, isManagedColumns]);
  var columnsWithStickyClasses = useMemo(function () {
    return columnsWithManageColumns.map(function (column, index) {
      return __assign(__assign({}, column), {
        className: classNames(column.className, isHorizontalScroll ? getStickyClass(columnsWithManageColumns, index) : null)
      });
    });
  }, [columnsWithManageColumns, isHorizontalScroll, isHorizontalScroll]);
  useRunOnceWhenTruthy(function () {
    defaultSelectedRows === null || defaultSelectedRows === void 0 ? void 0 : defaultSelectedRows.forEach(function (rowId) {
      toggleRowSelection(rowId)(true);
    });
  }, !!defaultSelectedRows && !!table.selectionState);
  return useMemo(function () {
    return {
      rows: processedRows,
      organizedRows: organizedRows,
      allColumns: columns,
      columns: columnsWithStickyClasses,
      sortState: table.sortState,
      filterState: table.filterState,
      rowsState: table.rowsState,
      selectionState: table.selectionState,
      toggleSort: toggleSort,
      updateFilterState: updateFilterState,
      updateRowState: updateRowState,
      updateTextFilter: updateTextFilter,
      toggleRowSelection: toggleRowSelection,
      selectRows: selectRows,
      resetFilters: resetFilters,
      pagination: pagination,
      pageSize: pageSize,
      updateColumnState: updateColumnState,
      isHorizontalScroll: isHorizontalScroll,
      columnsState: table.columnsState,
      isLazyLoading: isLazyLoading
    };
  }, [table, organizedRows, toggleSort, updateFilterState, updateRowState, updateTextFilter, toggleRowSelection, selectRows, processedRows, resetFilters, pagination, columnsWithManageColumns, updateColumnState, isLazyLoading]);
};

export { getFilterOptions, useTable };
//# sourceMappingURL=useTable.js.map
